/*------------------------------------------------------------------------------
  TREPT23.PRG  - adaptation of REPT23.PRG by Roger Seiler: roger@leadersoft.com
     Printing .TXT files or cText memvar passed from report creation .prg.
     Also includes specifying font by field and underlaying of boxes, shadows,
     lines, and bitmaps. See explanatory note at bottom of this file.
-------------------------------------------------------------------------------*/

/* Adapt this program to print text file output from other report generators
   that create Clipper DOS code such as Bandit, R&R Code Generator, or if just 
   directly coded in Clipper.  To "instantly" convert to FiveWin any such  
   pre-existing Clipper DOS reports, just re-program the lead-in screen 
   prompts in FiveWin. Then tack on the Clipper DOS procedural code that 
   actually creates the report - BUT be sure to have the report output to 
   disk as a .TXT file. Then use this .PRG to get FiveWin to print the .TXT 
   file or report text memvar passed as a parameter.

   A full explanation of the special codes for fonts and graphics is at the
   bottom of this file. Also info on how to access this from a program.

   To aid in the conversion of hard coded Clipper reports to create output
   that can be printed with Trept23, use ToText.prg accompanying this prg.

   IMPORTANT: Make sure that PREVIEW.DLL (from \fivewin\samples\report or
   from \fivewin\dll subdir) has been copied into your application
   subdirectory so that the Report Engine can have access to the dialog
   resources it needs when running in your application.  This .dll must be
   distributed with your ap.  Otherwise you'll get an error: CAN'T CREATE
   DIALOG etc. With FWH, use PREV32.DLL for 32 bits.

   Also note: a dummy function GetIniNam() is included and commented out at 
   about ln 1034 in case the app's main proc doesn't have such a func. In
   that case, uncomment this proc here to avoid linking error.

*/ 

#include "FiveWin.ch"
#include "report.ch"
// CHAR_PATTERN in Report.ch has to be changed from "B" to "b" if using
// proportional spaced fonts, or else will get "report wider than page" error.


*-----------------
CLASS TRept23

DATA aBoxSpecs,aFonts,aSpool

DATA cCurdir,cIniFile,cIniFileW,cIniPrefx,cLpt1,cName,cPrinter,cRefDir,;
     cText,cTextfile

DATA lLpt1,lPreview,lPrnstat,lPrntFontN

DATA nBoxCor,nFor,nLines,nLmargin,nPage,nPass,nRecno

DATA oDevice,oFont1,oFont2,oFont3,oFont4,oFont5,oFont6,oFont7,oReport

*---
     
METHOD New(cTextfile,cText,aFonts,aSpool,lPreview,nLmargin)  

METHOD Activate()

METHOD BoxSpecs()

  // Print bitmaps, boxes, lines and shadows, using ::aBoxSpecs data...
METHOD BoxWork()

METHOD GetTop()

METHOD SayMemo()


ENDCLASS

*-----------------

METHOD New(cTextfile,cText,aFonts,aSpool,lPreview,nLmargin) CLASS TRept23
LOCAL cIniSectn := ""
                           // The report, which was output to a text file
                           // by some report program, gets passed here
                           // as cTextfile.
                           // Alternatively, cText allows printing from
                           // a text memvar instead of a text file. 
                           // An array of 6 fonts can be passed, which will
                           // be utilized via special font codes imbedded
                           // in the text similar to HTML codes.
                           // aSpool is for use in 16 bits where there is
                           // a 64K memvar limit - it passes an array of
                           // names of text files into which the report has
                           // been divided.
                           // lPreview indicates if output is to go to
                           // the preview screen of TRpreview.
                           // nLmargin is an optional param indicating
                           // the number of pixels at which to set the left
                           // margin, overriding the regular setting.
                        
  DEFAULT aFonts := {{,,,},{,,,},{,,,},{,,,},{,,,},{,,,}}
  DEFAULT aSpool := {}
  DEFAULT lPreview := .f.
  DEFAULT nLmargin := 0

  ::aBoxSpecs := {}
  ::aFonts := aFonts
  ::aSpool := aSpool
  ::cCurdir := ""
  ::cIniFile := ""
  ::cIniFileW := ""
  ::cLpt1 := ""
  ::cText := cText
  ::cTextfile := cTextfile
  ::cName := "Report"
  ::cPrinter := GetProfString("Windows","Device","DeskJet")
  ::cRefDir := "" 
  ::lLpt1 := .t.
  ::lPreview := lPreview
  ::lPrnstat := .t.
  ::lPrntFontN := .f.
  ::nBoxCor := 0
  ::nFor := 0
  ::nLines := 0
  ::nLmargin := nLmargin
  ::nPage := 0
  ::nPass := 0
  ::nRecno := 0

  IF LEN(::aFonts) < 6
     AADD(::aFonts,{,,,})
     AADD(::aFonts,{,,,})
  ENDIF

  ::cRefDir := GetRefDir()  // Call to main prg to get main dir name

   IF !EMPTY(::cRefDir)
      ::cCurdir := UPPER(ALLTRIM(::cRefDir))
     ELSE
      ::cCurdir := CurDir()
      IF RAT("\",::cCurdir) > 0
         ::cCurdir := SUBSTR(::cCurdir, RAT("\",::cCurdir)+1,10)
         ::cCurdir := Alltrim(::cCurdir)
      ENDIF
   ENDIF

   ::cIniPrefx := ""   // In case ini name not same as main dir name.
   ::cIniPrefx := GETIniNam() // If this func not in main .prg then uncomment
                              // it at bottom of this file.

   IF EMPTY(::cIniPrefx)
      ::cIniFile  := CurDrive()+":\"+CurDir()+"\"+::cCurDir+".ini"
      ::cIniFileW := GetWinDir()+"\"+::cCurDir+".ini"
   ELSE
      ::cIniFile  := CurDrive()+":\"+CurDir()+"\"+::cIniPrefx+".ini"
      ::cIniFileW := GetWinDir()+"\"+::cIniPrefx+".ini"
   ENDIF

   IF FILE(::cIniFileW)
*      ::cIniFile := ::cIniFileW
   ENDIF

   IF !EMPTY(::cIniPrefx)
      cIniSectn := ::cIniPrefx
   ELSE
      cIniSectn := ::cCurDir
   ENDIF
   IF EMPTY(::nBoxCor)
      ::nBoxCor := VAL(GetPvProfString( cIniSectn , "BoxCor" , ".0" , ::cIniFileW ))
      ::cLpt1   := GetPvProfString( cIniSectn , "Lpt1" , "T" , ::cIniFileW )
   ENDIF
   IF ::cLpt1 = "T"
      ::lLpt1 := .t.
     ELSE
      ::lLpt1 := .f.
   ENDIF

  ::aBoxSpecs := {}
  ::nPage := 0
  ::nFor := 0
  ::nLines := 0
  ::nPass := 0

  SysRefresh()

  IF EMPTY(::cText) .AND. !EMPTY(::cTextfile)
     ::cText := OemToAnsi(MEMOREAD((::cTextfile)))
  ENDIF

RETURN Self
*------------------------
METHOD Activate() CLASS TRept23
LOCAL oObject

oObject := Self

  // Parse cText searching for codes requiring bitmaps, boxes, shadows or
  // lines, and then load those specs into ::aBoxSpecs
  ::aBoxSpecs := ::BoxSpecs()    // ***********

  IF ::lLpt1 .AND. Isprinter() = .f. // A Clipper function that checks if Lpt1 is ready.
        ::lPrnstat := MsgYesNo("Printer not connected on Lpt1." + CRLF + ;
            "If printer connected on other than Lpt1, reply YES to continue.")
        IF ::lPrnstat
            // If user indicates using Lpt2, then set .INI so the question
            // won't be asked again...
           WritePProString(::cCurDir,"Lpt1" , "F", ::cIniFileW )
        ENDIF
  ENDIF

  IF ::lPrnstat

    IF !::lPreview
     // Call up Windows' printer specs. dialog:
     ::oDevice  := TPrinter():New(::cName,.T.)
     ::oDevice:SetPortrait()  // In case user didn't
    ENDIF


  // Define default font...
  IF AT("DESKJET",UPPER(::cPrinter)) > 0 .OR. ::lPrntFontN = .t.
     DEFINE FONT ::oFont1 NAME "COURIER NEW" SIZE 0,-12  // A "True Type" font.
    ELSE
     DEFINE FONT ::oFont1 NAME "COURIER" SIZE 0,-12    // Not a "True Type" font.
  ENDIF
  ACTIVATE FONT ::oFont1
            // Font sizes for PRINTING are given in "points" not pixels.
            // (Pixels are used for font width and height for SCREENS only.)
            // By using a "0" for width and a "-" in front of the "height" #
            // we tell FiveWin that we're using points instead of pixels.

            // Thus -12 means "12 point," which is used for "10 pitch"
            // non-proportionally spaced Courier type that prints at
            // 6 lines/inch.

            // -10 means a "10 point" or "12 pitch" equal-spaced Courier font 
            // printing at 7 lines/inch.

            // Use -9 for a "15 pitch" equal-spaced Courier font printing
            // at 8 lines per inch.

   // Define up to 6 fonts passed in via ::aFonts from calling proc...

   // (Had to do these in nearly reverse order as workaround due to odd incorrect 
   // assignment of bold to oFont2 when oFont3 is supposed to be set to bold.)
   IF !EMPTY(::aFonts[2,1])
     DEFINE FONT ::oFont3 NAME ::aFonts[2,1] SIZE ::aFonts[2,2],::aFonts[2,3]
     IF(::aFonts[2,4] = "BOLD", ::oFont3:lBold := .t.,::oFont3:lBold := .f.)
     IF(::aFonts[2,4] = "ITALIC", ::oFont3:lItalic := .t.,)
     IF(::aFonts[2,4] = "UNDERLINE", ::oFont3:lUnderline := .t.,)
    ELSE
     DEFINE FONT ::oFont3 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
   ACTIVATE FONT ::oFont3

   IF !EMPTY(::aFonts[3,1])
     DEFINE FONT ::oFont4 NAME ::aFonts[3,1] SIZE ::aFonts[3,2],::aFonts[3,3]
     IF(::aFonts[3,4] = "BOLD", ::oFont4:lBold := .t.,::oFont4:lBold := .f.)
     IF(::aFonts[3,4] = "ITALIC", ::oFont4:lItalic := .t.,)
     IF(::aFonts[3,4] = "UNDERLINE", ::oFont4:lUnderline := .t.,)
    ELSE
     DEFINE FONT ::oFont4 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
   ACTIVATE FONT ::oFont4

   IF !EMPTY(::aFonts[1,1])
     DEFINE FONT ::oFont2 NAME ::aFonts[1,1] SIZE ::aFonts[1,2],::aFonts[1,3]
     // Temp...
     IF(::aFonts[1,4] = "BOLD", ::oFont2:lBold := .t.,::oFont2:lBold := .f.)
     IF(::aFonts[1,4] = "ITALIC", ::oFont2:lItalic := .t.,)
     IF(::aFonts[1,4] = "UNDERLINE", ::oFont2:lUnderline := .t.,)
    ELSE
     DEFINE FONT oObject:oFont2 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
   ACTIVATE FONT ::oFont2

   IF !EMPTY(::aFonts[4,1])
     DEFINE FONT ::oFont5 NAME ::aFonts[4,1] SIZE ::aFonts[4,2],::aFonts[4,3]
     IF(::aFonts[4,4] = "BOLD", ::oFont5:lBold := .t.,::oFont5:lBold := .f.)
     IF(::aFonts[4,4] = "ITALIC", ::oFont5:lItalic := .t.,)
     IF(::aFonts[4,4] = "UNDERLINE", ::oFont5:lUnderline := .t.,)
    ELSE
     DEFINE FONT ::oFont5 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
   ACTIVATE FONT ::oFont5
*cc
/*
msginfo(::oFont2:lBold,"::oFont2:lBold - 1")
msginfo(::oFont2:lItalic,"::oFont2:lItalic - 2")
msginfo(::oFont3:lBold,"::oFont3:lBold - 3")
msginfo(::oFont3:lItalic,"::oFont3:lItalic - 4")
msginfo(::oFont4:lBold,"::oFont4:lBold - 5")
msginfo(::oFont4:lItalic,"::oFont4:lItalic - 6")
msginfo(::oFont5:lBold,"::oFont5:lBold - 7")
msginfo(::oFont5:lItalic,"::oFont5:lItalic - 8")
*/
  IF LEN(::aFonts) > 4
   IF !EMPTY(::aFonts[5,1])
     DEFINE FONT ::oFont6 NAME ::aFonts[5,1] SIZE ::aFonts[5,2],::aFonts[5,3]
     IF(::aFonts[5,4] = "BOLD", ::oFont6:lBold := .t.,::oFont6:lBold := .f.)
     IF(::aFonts[5,4] = "ITALIC", ::oFont6:lItalic := .t.,)
     IF(::aFonts[5,4] = "UNDERLINE", ::oFont6:lUnderline := .t.,)
    ELSE
     DEFINE FONT ::oFont6 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
  ENDIF
  ACTIVATE FONT ::oFont6

  IF LEN(::aFonts) > 5
   IF !EMPTY(::aFonts[6,1])
     DEFINE FONT ::oFont7 NAME ::aFonts[6,1] SIZE ::aFonts[6,2],::aFonts[6,3]
     IF(::aFonts[6,4] = "BOLD", ::oFont7:lBold := .t.,::oFont7:lBold := .f.)
     IF(::aFonts[6,4] = "ITALIC", ::oFont7:lItalic := .t.,)
     IF(::aFonts[6,4] = "UNDERLINE", ::oFont7:lUnderline := .t.,)
    ELSE
     DEFINE FONT ::oFont7 NAME "COURIER NEW" SIZE 0,-12
   ENDIF
  ENDIF
  ACTIVATE FONT ::oFont7

            // Note that in creating the report object, we are leaving out
            // any mention of title or header specs. because those are
            // already included in our .TXT file.
  IF !::lPreview
     REPORT ::oReport ;
          FONT ::oFont1,::oFont2,::oFont3,::oFont4,::oFont5,::oFont6,::oFont7 ;
          CAPTION "Text Printout" ;
          TO DEVICE ::oDevice
    ELSE
     REPORT ::oReport ;
          FONT ::oFont1,::oFont2,::oFont3,::oFont4,::oFont5,::oFont6,::oFont7 ;
          CAPTION "Text Printout" ;
          PREVIEW
  ENDIF


     COLUMN DATA " " SIZE 53  // Was 76. Trick "Data" - we're fooling Mother 
                              // Nature here. The memo will go into the " ". 
                              // Changed from 76 to 53 to work with an unknown
                              // change in FWH 2.7.

     END REPORT

   IF ::oReport:lCreated = .f.      //::oDevice:hDC = 0
      ::oReport:End()
      ::oFont1:End();::oFont2:End();::oFont3:End();::oFont4:End();::oFont5:End()
      ::oFont6:End();::oFont7:End()
      RETURN ::lPrnstat
   ENDIF

             // All title lines, page numbers and headings already exist in
             // our text file, so the next two code statements make sure that
             // FiveWin won't add its own title lines.
     ::oReport:nTitleUpLine := RPT_NOLINE
     ::oReport:nTitleDnLine := RPT_NOLINE
     ::oReport:Margin(.15,RPT_LEFT,RPT_INCHES)
     ::oReport:Margin(.25,RPT_TOP,RPT_INCHES)
     ::oReport:Margin(.25,RPT_BOTTOM,RPT_INCHES)
     // I know the top/bottom margins shown here should mathematically cause
     // more than the desired 60 lines per page to print. However, in actual 
     // practice with various printers emulating the 
     // HP LaserJet Plus, this is what I had to use to get 60 lines.  Why?
     // Inscrutible printer driver!  Maybe your printer driver functions in
     // a more mathematically logical fashion.  If so, adjust accordingly.
     // The left margin is set so small here because a margin is already
     // built into the text file itself (in this particular case).
     IF !EMPTY(::nLmargin)
        ::oReport:Cargo := ::nLmargin // To override calc of left margin.
     ENDIF

     ACTIVATE REPORT oObject:oReport ;
          ON STARTPAGE oObject:BoxWork() ;
          ON INIT oObject:SayMemo() ;
          ON ENDPAGE ( IF( oObject:nFor = (oObject:nLines+1),;
                           (oObject:oReport:lBreak := .T.,;
                            oObject:oReport:oRptWnd:End();
                            ),;
                          );
                      )

     ::oFont1:End();::oFont2:End();::oFont3:End();::oFont4:End();::oFont5:End()
     ::oFont6:End();::oFont7:End() 
     ::oReport := NIL


  ENDIF //::lPrnstat

RETURN ::lPrnstat

*------------------------------------------------------

METHOD SayMemo() CLASS Trept23
// Problem 2/14/02: Between spool files, an extra blank line is printed.
// Tried to solve with decrementing nPageln, but no work.

   LOCAL cLine := "", cLpart := "",cCode := "",nAt := 0,nEnd := 0,nCol := 0,;
         nPageln := 0,nAt2 := 0,nEnd2 := 0,nLen := 0,nCnt := 0,nFont := 1,;
         nEndTxt := 0,nStart := 0,oFont,nCodLen,nSpace := 0,cSpace := "",;
         nTextLen := 0,nTcnt := 0,lEnd := .f.,nRepEnd := 0,lP := .f.,;
         nPad := 6,lTest := .f.,lFonts := .f.,nCnt2 := 0,;
         lStartsp := .f.,nX := 0,oDevice,hDc
   LOCAL nMultPoint := 0,cLine2 := "",nAt3 := 0,cLineWk := "",lChk := .f.,;
         nLen2 := 0,nAtb := 0,nAtc := 0,nAte := 0,nAts := 0,nAtl := 0,;
         nAtg := 0,nAtt := 0,nAth := 0,nLen3 := 0,;
         nCnt3 := 1,nCntBk := 1,nCnt4 := 1,aSet := {},nFor,nLines

   // nPad is a numeric code sent to Report:Say() to tell it what to do
   // with the printhead. 1 = justify left; 2 = justify right; 3 = justify center;
   // 4 = go top; 5 = go bottom; 6 = line continuation; 7 = position printhead at:

   IF LEN(::aSpool) > 0  // whether spool files are used for a large report.
      nCntBk := nLen3 := LEN(::aSpool)
      IF FILE(::aSpool[nCnt3])
         ::cText := OemToAnsi(MEMOREAD(::aSpool[nCnt3]))
            // Get rid of end of file marker ^Z to avoid printing a blank line
            // at end of spool...
         nAt := AT("<endsp>",::cText)-1  // Find end of spool text marker.
         IF nAt < 1
            nAt := LEN(::cText)
         ENDIF
         ::cText := SUBSTR(::cText,1,nAt)
         nAt := 0
         nCnt3++
        ELSE
         MsgAlert("Error: spool file not on disk.")
         RETURN nil
      ENDIF
   ENDIF

   ::cCurDir := CurDir()
   IF RAT("\",::cCurDir) > 0
      ::cCurDir := SUBSTR(::cCurDir, RAT("\",::cCurDir)+1,10)
      ::cCurDir := Alltrim(::cCurDir)
   ENDIF

   IF UPPER(::cCurDir) == "QCS360" .OR. UPPER(::cCurDir) == "QCSRV360" ;
      .OR. UPPER(::cCurDir) == "QCSWIN" .OR. UPPER(::cCurDir) == "S360" ;
      .OR. UPPER(::cCurDir) == "QCS360X"

      aSet    := GetaSet()
      ::cCurDir := aSet[27]

   ENDIF


   IF EMPTY(nMultPoint)
      nMultPoint := VAL(GetPvProfString( ::cCurDir , "MultPoint" , "1" , ::cIniFileW ))
   ENDIF

   CursorWait()

  DO WHILE nCntBk > 0  // For spool printing of large documents, nCntBk 
                       // counts backward as spool files are printed
                       // until the last one has been printed.
     nTextLen := LEN(::cText)
     ::nLines := MlCount(::cText,76) // Orig. text has up to 76 chars. per line.

    IF nCntBk = nLen3  // Only set to 0 initially.
     nPageln := 0
    ENDIF
*     ::oReport:BackLine(1)     // Not needed here. If used, will cause printing
                              // of subsequent pages to get out of synch by
                              // pushing the start of printing down 1 line
                              // per page.

   IF AT("<f", ::cText) > 0
      lFonts := .t.
     ELSE
      lFonts := .f.
   ENDIF

   nFor := ::nFor
   nLines := ::nLines
  FOR nFor := 1 TO nLines

                      // Must grab extra line width to include font codes...
    IF lFonts
       cLine := MemoLine(::cText, 250, nFor)
      ELSE
       cLine := MemoLine(::cText, 76, nFor)
    ENDIF
       // Make sure that on transition from one spool file to the next one,
       // that a bunch of blank pages don't get printed, due to nFor being
       // greater than the actual number of lines of text (due to use of 
       // font codes and grabbing 250 chars at a time instead of 76)...
    IF cLine == ""
       EXIT   // Don't print blank pages between spool files.
    ENDIF
    nLen := LEN(cLine)
          // The next 4 lines handle situation where new page follows a
          // short page with less than 60 lines. This is triggered
          // by "Page No."  Only works if "Page No." appears on
          // top left of page or top right of page.
    IF nFor > 3 .AND. ( AT("Page No.", cLine ) > 0 .AND.;
                        AT("Page No.", cLine ) < 90 )
          ::oReport:EndPage()
          nPageln := 0
    ENDIF

    IF lEnd
       ::oReport:lFinish := .t.
       IF !::lPreview
*          ::oReport:EndPage()  // This is handled in Report.prg
       ENDIF
       nPageln := 0
       EXIT
    ENDIF

    ::oReport:StartLine()

    cLineWk := cLine
    nLen2 := LEN(cLineWk)
    cLine2 := ""
       // Remove box/graphic codes from line's text...
    DO WHILE nLen2 > 0 .AND. nCnt2 < 8  
                    nAtb := AT("<b", cLineWk)  // box
                    nAtc := AT("<c", cLineWk)  // shaded box
                    nAte := AT("<e>", cLineWk) // extension of box to include variable length text
                    nAth := AT("<h>", cLineWk) // extension of line to include variable length text
                    nAts := AT("<s", cLineWk)  // shade
                    nAtl := AT("<l", cLineWk)  // line
                    nAtg := AT("<g", cLineWk)  // bitmap Graphic
                    nAtt := AT("<t", cLineWk)  // global indicator

       IF nAtb > 0 .AND. nAtc > 0
          nAt3 := MIN(nAtb, nAtc)
         ELSE
          nAt3 := MAX(nAtb, nAtc)
       ENDIF
       IF nAt3 > 0 .AND. nAte > 0
          nAt3 := MIN(nAt3, nAte)
         ELSE
          nAt3 := MAX(nAt3, nAte)
       ENDIF
       IF nAt3 > 0 .AND. nAts > 0
          nAt3 := MIN(nAt3, nAts)
         ELSE
          nAt3 := MAX(nAt3, nAts)
       ENDIF
       IF nAt3 > 0 .AND. nAtl > 0
          nAt3 := MIN(nAt3, nAtl)
         ELSE
          nAt3 := MAX(nAt3, nAtl)
       ENDIF
       IF nAt3 > 0 .AND. nAth > 0
          nAt3 := MIN(nAt3, nAth)
         ELSE
          nAt3 := MAX(nAt3, nAth)
       ENDIF
       IF nAt3 > 0 .AND. nAtg > 0
          nAt3 := MIN(nAt3, nAtg)
         ELSE
          nAt3 := MAX(nAt3, nAtg)
       ENDIF
       IF nAt3 > 0 .AND. nAtt > 0
          nAt3 := MIN(nAt3, nAtt)
         ELSE
          nAt3 := MAX(nAt3, nAtt)
       ENDIF

       IF nAt3 = 0
          lChk := .f.
          cLine := cLine2 + cLineWk
          EXIT
         ELSE
          lChk := .t.
       ENDIF

       IF lChk // Graphics used
          cLine2 := cLine2 + SUBSTR(cLineWk, 1, nAt3-1)  // 1st part of line
          cLineWk := SUBSTR(cLineWk, nAt3, LEN(cLineWk)) // Next part of line
          cLineWk := SUBSTR(cLineWk, AT(">", cLineWk)+1, LEN(cLineWk)) // Remainder of line minus graphic code

       ENDIF
       nLen2 := LEN(cLineWk)
       nCnt2++
       SysRefresh()
    ENDDO
    nCnt2 := 0

//  Handle assignment and changes of text fonts...
    nCol := 1
    nCnt := 1

    DO WHILE nLen > 0   
      nAt := AT("<f", cLine) // 1st font designation in the line.

      IF nAt > 0 
         nStart := nMultPoint * 250   // again below at line 303
         // Must first SAY() any part of line prior to "<f" without font spec. (because using prior font).
         IF nAt > 1 .AND. nCnt = 1 // If no fontcode at beginning of line, 
                                  // but do this only on first pass (nCnt = 1)!
            cLpart := SUBSTR(cLine,1,nAt-1)

            // METHOD Say(nCol, xText, nFont, nPad, nRow, nStart) CLASS TReport
            ::oReport:Say(1,cLpart,,,,nStart)  //  *** PRINT THE TEXT ***
            nCnt++
         ENDIF
         nEnd := AT(">", cLine)
         IF nEnd = 0
            MsgAlert('Error: ">" code missing at end of font spec.'+CRLF+;
                     cLine)
            EXIT
         ENDIF
               
         cCode := SUBSTR(cLine,nAt+1,nEnd-nAt-1)
         nFont := 1
         nPad := 6
         nCodLen := LEN(cCode) // Fonts in ::aFonts are all in addition to the 
                               // default font (set at line 118), hence
                               // aFont[1] is identified as nFont = 2...
         IF(SUBSTR(cCode,2,1) == "1",nFont := 2,)
         IF(SUBSTR(cCode,2,1) == "2",nFont := 3,)
         IF(SUBSTR(cCode,2,1) == "3",nFont := 4,)
         IF(SUBSTR(cCode,2,1) == "4",nFont := 5,)
         IF(SUBSTR(cCode,2,1) == "5",nFont := 6,)
         IF(SUBSTR(cCode,2,1) == "6",nFont := 7,)
         IF nCodLen > 2 .AND. nCodLen < 6
            IF(SUBSTR(cCode,4,1) == "s",nSpace := VAL(SUBSTR(cCode,5,1)),)
         ENDIF
         IF nCodLen = 6
            IF(SUBSTR(cCode,4,1) == "s",nSpace := VAL(SUBSTR(cCode,5,2)),)
         ENDIF
         IF nCodLen >= 6  // Handle tabbing over to a set printer head position
            IF(SUBSTR(cCode,4,1) == "p",lP := .t.,lP := .f.)
            IF(lP,nStart := VAL(SUBSTR(cCode,5,4)),)
            nStart := INT(nStart * nMultPoint)
              // Change nPad to 7 = "position printhead at" if cCode = "p"
            IF(lP,nPad := 7,nPad := 6)
         ENDIF
         cSpace := SPACE(nSpace)
         cLine := SUBSTR(cLine,nEnd+1,nLen)
         cLine := cSpace + cLine
         nLen := LEN(cLine)
         nAt2 := AT("<f", cLine) // next font spec in the line.
         IF nAt2 = 0
            nEndTxt := nLen
           ELSE
            nEndTxt := nAt2-1
         ENDIF
         cLpart := SUBSTR(cLine,1,nEndTxt)

         nRepEnd := AT("<end>", cLpart)
         IF nRepEnd > 0
            lEnd := .t.
            cLpart := SUBSTR(cLpart,1,nRepEnd-1)
         ENDIF
         IF nCnt = 1  // If fontcode comes at beginning of line...

          // METHOD Say(nCol, xText,nFont,nPad,nRow,nStart) CLASS TReport
            ::oReport:Say(nCol,cLpart,nFont,1,,nStart)  //  *** PRINT THE TEXT ***
           ELSE  // Most of the work done here...nPad usually = 6 = line continuation.
            ::oReport:Say(nCol,cLpart,nFont,nPad,,nStart)  //  *** PRINT THE TEXT ***
         ENDIF
         nCnt++
         nSpace := 0
         cSpace := ""
         nStart := nMultPoint * 250   // reset to line start pos. after tabbing
         IF nAt2 > 0
            cLine := SUBSTR(cLine,nAt2,nLen)
            nLen := LEN(cLine)
          ELSE
            nLen := 0
         ENDIF

       ELSE  // If on 1st pass, line has no font codes...
             // Defaults to last font used...
         nRepEnd := AT("<end>", cLine)
             // If getting an extra blank page ejected at end of report,
             // then put "<end>" at end of text string and this will
             // set ::oReport:lFinish to .t. so that another page won't eject.
         IF nRepEnd > 0
            lEnd := .t.
            ::oReport:lFinish := .t.
            cLine := SUBSTR(cLine,1,nRepEnd-1)
         ENDIF
         ::oReport:Say(1,cLine,nFont,,,nStart)  //  *** PRINT THE TEXT ***
         nLen := 0
      ENDIF  // nAt > 0 regarding if font codes used
      IF nCnt > 20  // Emergency escape.
         EXIT
      ENDIF
      SysRefresh()
      CursorWait()
    ENDDO  // nLen > 0  where nLen is length of cLine.

    ::oReport:EndLine()

             // The following section checks to see if we've reached the
             // bottom of the page (60 lines). If so, it then calls
             // ::GetTop() to eliminate empty spaces in the text file between
             // the bottom of the outgoing page and the top line of the
             // next page.  The intent of this code is to get rid of the
             // empty lines that most report generators insert between pages
             // when output is to a text file -- otherwise our page printing
             // would get out of synch with the formating of the original
             // report. This approach only works where printing (including the
             // page number and other page heading data) is supposed to appear
             // at the same place at the top of every page.  
          nPageln := nPageln + 1
          IF nPageln = 60
             ::nFor := nFor
             nFor := ::GetTop()
             nPageln := 0
          ENDIF

          SysRefresh()
          CursorWait()
   NEXT   // FOR nFor := 1 TO nLines
          // Erase spool files after they've been printed...
   IF LEN(::aSpool) > 0 
      nCnt4 := IF(nCnt3 > 1, nCnt3-1, 1)
      IF FILE(::aSpool[nCnt4])
         ERASE (::aSpool[nCnt4])
      ENDIF
   ENDIF
   nCntBk--
   IF nCnt3 <= nLen3
     IF FILE(::aSpool[nCnt3])
         ::cText := OemToAnsi(MEMOREAD(::aSpool[nCnt3]))
             // Get rid of end of spool file marker...
         nAt := AT("<endsp>",::cText)-1  
         IF nAt < 1
            nAt := LEN(::cText)
         ENDIF
         ::cText := SUBSTR(::cText,1,nAt)
         nAt := 0
         nCnt3++
            // Back up one row, because blank row was added at end of
            // last spool file...
         ::oReport:nRow := ::oReport:nRow - ::oReport:nStdLineHeight
        ELSE
         MsgAlert("Error: spool file not on disk.")
         RETURN nil
     ENDIF
   ENDIF
   SysRefresh()
   CursorWait()
  ENDDO  // nCntBk > 0

*    ::oReport:Newline()  // Only activate this line if you need to have the
                        // program eject another page upon completing printing.
    IF !::lPreview
     // Call up Windows' printer specs. dialog:
     ::oReport:oDevice:ResetDC()

    ENDIF

RETURN NIL

*----------------------

METHOD GetTop() CLASS TRept23
LOCAL lTest := .T.,cLine := "",nFor := 0,nLines := 0
nFor := ::nFor
nLines := ::nLines
DO WHILE lTest = .T. .AND. nFor <= nLines
   nFor := nFor + 1
   cLine := MemoLine(::cText, 76, nFor)
   lTest := EMPTY(cLine)
   SysRefresh()   // Inserted to allow Windows to catch up with housekeeping
                  // that is held up by the "DO WHILE."
ENDDO
nFor := nFor - 1
RETURN nFor
*----------------------

METHOD BoxSpecs() CLASS TRept23
LOCAL aBack := {},nLine := 0,lChk := .f.,cLine := "",nAt := 0,i := 0,;
      nPg := 0,cCode := "",cTemp := "",nTop := 0,nLeft := 0,nLen := 0,;
      nBot := 0,nRt := 0,nPenW := 0,cColor := "",cBitmap := "",nLastB := 0,;
      nCnt := 0,nTst := 0,nPgChk := 0,nPgChk2 := 0,nBoxRow := 0,nLastL := 0,;
      nNewBot := 0,cTExt := ""

cText := ::cText

::nPass++
aBack := ARRAY(1,10)
aBack[1,1] := 0
aBack[1,2] := 0
aBack[1,3] := ""
aBack[1,4] := 0
aBack[1,5] := 0
aBack[1,6] := 0
aBack[1,7] := 0
aBack[1,8] := 0
aBack[1,9] := ""
aBack[1,10] := ""

IF AT("<t>", cText) = 0  // If graphics used in report, insert a <t> code in
   RETURN aBack          // the 1st page header text. Otherwise this process
ENDIF                    // will abort to save processing time.

IF AT("<b", cText) > 0 .OR.;   // box
   AT("<c", cText) > 0 .OR.;   // shaded box
   AT("<e>", cText) > 0 .OR.;  // extension of box to include variable length text
   AT("<h>", cText) > 0 .OR.;  // extension of line to include variable length text
   AT("<s", cText) > 0 .OR.;   // shade
   AT("<l", cText) > 0 .OR.;   // line
   AT("<g", cText) > 0         // bitmap Graphic

   lChk := .T.
ENDIF
IF !lChk // Graphics not used
   RETURN aBack
ENDIF   

/*
The aBack array row:
1. The page #
2. The row # - based on a count of CHR(13)+CHR(10) codes
3. The action code - B (box), C (box w/shadow), S (shadow), L (line), G (bitmap graphic)
4. From the row # at left edge of paper, the nTop in inches
5. From the row # at left edge of paper, the nLeft in inches
6. From the row # at left edge of paper, the nBottom in inches
7. From the row # at left edge of paper, the nRight in inches
8. The thickness of the pen to be used, if any.
9. The color for oBrush, if any.
10. The name of the bitmap, if any.
*/

DO WHILE LEN(cText) > 5

   // Find end of text line by finding carriage return...
 nAt := AT(CHR(13)+CHR(10), cText)
 IF nAt > 0
   IF nAt > 1
      // Load line to be analyzed...
      cLine := SUBSTR(cText, 1, nAt-1)
      nCnt := 0
     ELSE
      cLine := ""
   ENDIF
   nLen := LEN(cLine)
    // Determine if new line also signals start of a new page...
    IF ( AT("Page No.", cLine ) > 0 .AND.;
         AT("Page No.", cLine ) < 90 ) .OR.;
           (::nPass = 1 .AND. nLine = 0)
      nPg++
      nLine := 0
   ENDIF
      // Reset cText to remainder of text after the line being processed...
   cText := SUBSTR(cText, nAt+2, LEN(cText))
   nLine++
  ELSE
   EXIT
 ENDIF

 // Layout of graphic Codeset...
 //    top  left bot  rt   th brushclr bmp - need 6 commas
 // <b 1.33,1.33,1.33,1.33,1,CLR_HGRAY,BLACKINK.BMP>

 // See if line has a graphics codeset...
             nAt := AT("<b", cLine)
 IF(nAt = 0, nAt := AT("<c", cLine),)
 IF(nAt = 0, nAt := AT("<e>", cLine),)
 IF(nAt = 0, nAt := AT("<h>", cLine),)
 IF(nAt = 0, nAt := AT("<s", cLine),)
 IF(nAt = 0, nAt := AT("<g", cLine),)
 IF(nAt = 0, nAt := AT("<l", cLine),)

 DO WHILE nAt > 0   // the line has a graphics codeset...
  cLine := SUBSTR(cLine,nAt,LEN(cLine))
  IF SUBSTR(cLine,2,1) == "e"
    IF nLine # aBack[nLastB,2]// Avoid box extension if still on starting line.
     // Update box nBot for <e>
     IF nPgChk # nLastB
        nBoxRow := aBack[nLastB,6]
     ENDIF
     nPgChk := nLastB
     nNewBot := nBoxRow + ((nLine-aBack[nLastB,2]) / 6) 
      // Make sure new nBot isn't the top of the next page (a lesser number
      // than prior line), or else weird things will happen!...
     nNewBot := MAX(nNewBot,nBoxRow)
     aBack[nLastB,6] := nNewBot
    ENDIF
   ELSEIF SUBSTR(cLine,2,1) == "h"
    IF nLine # aBack[nLastL,2] // Avoid line extension if still on starting line.
     // Update line nBot for <h>
     IF nPgChk2 # nLastL
        nBoxRow := aBack[nLastL,6]
     ENDIF
     nPgChk2 := nLastL
     aBack[nLastL,6] := nBoxRow + ((nLine-aBack[nLastL,2]) / 6) 
    ENDIF
   ELSE
    // Increment array row number (each graphics code has its own row)...
    i++
    IF !EMPTY(aBack[LEN(aBack),1])  // Except first pass, add another empty row...
       AADD(aBack, ARRAY(10))
    ENDIF
    aBack[i,1] := nPg

    aBack[i,2] := nLine
    cCode := SUBSTR(cLine,2,55)
    aBack[i,3] := SUBSTR(cCode,1,1)            // Action code
    IF aBack[i,3] = "b" .OR. aBack[i,3] = "c"
      nLastB := i
    ENDIF
    IF aBack[i,3] = "l" 
      nLastL := i
    ENDIF

    cCode := SUBSTR(cCode,1,AT(">",cCode)-1)
    cTemp := SUBSTR(cCode,3,AT(",",cCode)-3)
    IF(!EMPTY(cTemp),nTop := VAL(cTemp),nTop := 0)    // nTop

    cCode := SUBSTR(cCode,AT(",",cCode)+1,55)
    cTemp := SUBSTR(cCode,1,AT(",",cCode)-1)
    IF(!EMPTY(cTemp),nLeft := VAL(cTemp),nLeft := 0)  // nLeft

    cCode := SUBSTR(cCode,AT(",",cCode)+1,55)
    cTemp := SUBSTR(cCode,1,AT(",",cCode)-1)
    IF(!EMPTY(cTemp),nBot := VAL(cTemp),nBot := 0)    // nBot

    cCode := SUBSTR(cCode,AT(",",cCode)+1,55)
    cTemp := SUBSTR(cCode,1,AT(",",cCode)-1)
    IF(!EMPTY(cTemp),nRt := VAL(cTemp),nRt := 0)      // nRt

    cCode := SUBSTR(cCode,AT(",",cCode)+1,55)
    cTemp := SUBSTR(cCode,1,AT(",",cCode)-1)
    IF(!EMPTY(cTemp),nPenW := VAL(cTemp),nPenW := 0)  // nPenW

    cCode := SUBSTR(cCode,AT(",",cCode)+1,55)
    cTemp := SUBSTR(cCode,1,AT(",",cCode)-1)
    IF(!EMPTY(cTemp),cColor := cTemp,cColor := "")    // cColor for oBrush

    cTemp := SUBSTR(cCode,AT(",",cCode)+1,55)
    IF(!EMPTY(cTemp),cBitmap := cTemp,cBitmap := "")  // cBitmap

    // Now load the rest of the array row data...
    aBack[i,4] := nTop
    aBack[i,5] := nLeft
    aBack[i,6] := nBot
    aBack[i,7] := nRt
    aBack[i,8] := nPenW
    aBack[i,9] := cColor
    aBack[i,10] := cBitmap
  ENDIF // If <e>
  // See if remainder of line has another graphics codeset...
  cLine := SUBSTR(cLine,AT(">", cLine)+1,LEN(cLine))

              nAt := AT("<b", cLine)
  IF(nAt = 0, nAt := AT("<c", cLine),)
  IF(nAt = 0, nAt := AT("<e>", cLine),)
  IF(nAt = 0, nAt := AT("<h>", cLine),)
  IF(nAt = 0, nAt := AT("<s", cLine),)
  IF(nAt = 0, nAt := AT("<g", cLine),)
  IF(nAt = 0, nAt := AT("<l", cLine),)
  nCnt++
  IF nCnt > 7  // Safety bailout to prevent endless looping.
     EXIT
  ENDIF
  SysRefresh()
 ENDDO  // nAt = 1
 SysRefresh()
ENDDO

RETURN aBack
*----------------------
  // Print bitmaps, boxes, lines and shadows, using ::aBoxSpecs data...
METHOD BoxWork() CLASS TRept23
LOCAL cCode := "",i := 0,n := 0,nSeek := 1
LOCAL nInchWidth, nInchHeight,nTst := 0
LOCAL oBrush,nRow := .1,nCol := .1,nBottom := 1,nRight := 1,aStart := {},aEnd := {}
LOCAL nTop := .1,nLft := .1,nBot := .1,nRt := .1,nPenW := 1,nTmargin := .5

::nPage++

nInchWidth  := ::oReport:PhyWidth(RPT_INCHES)
nInchHeight := ::oReport:PhyHeight(RPT_INCHES)

/*
The aBack array row:
1. The page #
2. The row # - based on a count of CHR(13)+CHR(10) codes
3. The action code - B (box), C (box w/shadow), S (shadow), L (line), G (bitmap graphic)
4. From the row # at left edge of paper, the nTop in inches
5. From the row # at left edge of paper, the nLeft in inches
6. From the row # at left edge of paper, the nBottom in inches
7. From the row # at left edge of paper, the nRight in inches
8. The thickness of the pen to be used, if any.
9. The color for oBrush, if any.
10. The name of the bitmap, if any.
*/

  // Find first array element with codes for current page...

nSeek := ::nPage
AEVAL(::aBoxSpecs, {|row| n++, IF(row[1] = nSeek,(i := n,nSeek--),) })

IF i > 0
 DO WHILE ::aBoxSpecs[i,1] = ::nPage // Get array row and check its page no.

   // ::nBoxCor comes from .INI and is a decimal correction factor giving
   // the fraction of an inch by which the box/line/bmp is to be offset up and
   // to the right to correct for the printer's alignment peculiarities....
   // (Note: the -.2 was inserted below to correct for FWH 2.7 change...
  nTop := ::aBoxSpecs[i,4] + nTmargin + (::aBoxSpecs[i,2] / 6) - .2 - ::nBoxCor // assuming six lines per inch.
  nLft := ::aBoxSpecs[i,5] + ::nBoxCor
  nBot := ::aBoxSpecs[i,6] + nTmargin + (::aBoxSpecs[i,2] / 6) - .2 - ::nBoxCor // assuming six lines per inch.
  nRt  := ::aBoxSpecs[i,7] + ::nBoxCor
  nPenW := ::aBoxSpecs[i,8]
                                  // C = box with a shadow within. 
  IF UPPER(::aBoxSpecs[i,3]) == "B" .OR. UPPER(::aBoxSpecs[i,3]) == "C" 

     ::oReport:Box(nTop             ,;             // nRow
                 nLft             ,;             // nCol
                 nBot             ,;             // nBottom
                 nRt              ,;             // nRight
                 nPenW            ,;             // first pen created
                 RPT_INCHES)
  ENDIF

  IF UPPER(::aBoxSpecs[i,3]) == "L"   // Draw a Line
     ::oReport:Line(nTop    ,;                     // nTop
                  nLft    ,;                     // nLeft
                  nBot    ,;                     // nBottom
                  nRt     ,;                     // nRight
                  nPenW   ,;                     // second pen created
                  RPT_INCHES)
  ENDIF

  IF UPPER(::aBoxSpecs[i,3]) == "S" .OR. UPPER(::aBoxSpecs[i,3]) == "C"  // C = a box with a shadow
  // RWS's drawing of a shadow...

   IF oBrush == NIL
       DEFINE BRUSH oBrush COLOR ::oReport:nClrShadow
   ENDIF

      IF UPPER(::aBoxSpecs[i,3]) == "C"
         nTop := nTop + .01
         nLft := nLft + .01
         nBot := nBot - .01
         nRt  := nRt  - .01
      ENDIF


      aStart := ::oReport:oDevice:Inch2Pix(nTop, nLft)
      aEnd   := ::oReport:oDevice:Inch2Pix(nBot, nRt)
      ::oReport:oDevice:FillRect({aStart[1] ,;
                                aStart[2] ,;
                                aEnd[1]   ,;
                                aEnd[2] } ,;
                                oBrush)
  ENDIF // Shadow

  IF UPPER(::aBoxSpecs[i,3]) == "G" // bitmap Graphic

      // Must cancel ::nBoxCor correction for nBot...
     nBot := nBot + ::nBoxCor

     ::oReport:SayBitmap(nTop,nLft,UPPER(::aBoxSpecs[i,10]),nRt,nBot)

  ENDIF // bitmap Graphic

/*        
    These are the parameters of the method Saybitmap:
    
    1) nRow
    2) nCol
    3) cBitmap file (only)
    4) nWidth
    5) nHeight
    6) nScale
    
    If Scale is in INCHES (1) then 1,2,4 and 5 are in inches, and if Scale
    is in CMETERS (2) then 1,2,4 and 5 are in centimeters.
*/
  IF i + 1 > LEN(::aBoxSpecs)
     EXIT
  ENDIF
  i++  // To get next array row.
 ENDDO // ::nPage
ENDIF // > 0

RETURN NIL
*----------------------------
 // Activate this func. if it doesn't exist in app's main prg.
*STATIC FUNCTION GetIniNam() ; RETURN ""
*----------------------------

/*
Explanation of special codes for font assignment by field and also codes
for printing boxes, lines, shades and bitmaps.

FONTS

You can specify up to 6 different fonts to be used in a report. Pass them
to TREPT23():New() in an array, aFonts, in the following layout:

aFonts := {;
          {"TIMES NEW ROMAN", 0,-11, ""},;
          {"TIMES NEW ROMAN", 0,-11, "BOLD"},;
          {"TIMES NEW ROMAN", 0,-11, "ITALIC"},;
          {"", 0,0, ""},;
          {"ARIAL", 0,-14, "BOLD"},;
          {"ARIAL", 0,-24, "BOLD"};
          }

          Each array element has the font name, the width ("0"), the height
          (always a negative "points" number, and whether the font is
          standard (""), "BOLD" or "ITALIC". If the specs are blank, then
          the default Courier 0,-12 font is used, as specified in REPT23.

The call to TREPT23() from within your program is as follows:
                                                                 // Create obj.
    oRept23 := TRept23():New(cFile,cText,aFonts,aSpool,lPreview,nMargin)
    oRept23:Activate()                                           // Do printing
    oRept23 := nil                                               // Retire obj.

    Relevant Parameters:
      cFile  ...   If printing a report stored in a text file, 
                   the name of the file.
      cText  ...   If printing from a report created in a text string memory
                   variable, the name of that string memvar.
      aFonts ...   The array with the fonts used in the report, as above. 
      aSpool ...   If using in 16 bit FW app, and if report's memvar will
                   exceed 64k, then due to 64k memvar limit, pass in an array
                   with names of sequential text files, each containing a
                   sequential portion of the report.
      lPreview ... Whether output shall go to on-screen preview tool, RPreview.
      nMargin ...  Number of pixels from left edge of paper to force position
                   of printing, overriding regular left margin. (For graphics.)

In the cText or cFile text sent to REPT23, the indication of font codes is
done in a way similar to HTML program code, in that the font codes are 
contained within < > symbols. Example: <f1> which means to use the first font
in the fonts array. The font code must be placed in front of the field
of text to which it applies. The space taken by the font code does not count
as part of the text length of the line, so if your line length is set at
70 characters, and you have a line with 69 characters plus a 3 character font
code at the beginning, it just counts as a 69 character line as far as margins
and printed line length are concerned.

Using the fonts specified in the array above, you could specify fonts
for various fields as shown in this example of text to be printed:

   <f2>My Favorite Poem... <f3 s4>by Anonymous Artist

   <f1>Mary had a little lamb
            whose fleece were white as snow.

            <f3 p450>To be continued...

In the text above, <f2> means that "My Favorite Poem..." will be printed in 
Times Roman Bold, 11 point type.  Next, <f3 s4> says that "by Anonymous Artist" 
will be printed in Times Roman Italic, also 11 point, and that the print head
is to move 4 extra spaces to the right before printing this field after
printing the previous field. This is to allow a way of adding compensating 
spaces when converting from Courier same-space characters (in which the
report may have been designed) to a proportional space type style like
Times Roman. So if converting to a proportional space font (which bunches up
blank spaces into much less space than same-space typewriter style Courier
font) causes there to be less space between two fields of text than you want, 
then use the spaces code "s" and the number of spaces you want added. Just 
insert this code right after the fint code within the same pointed brackets
as shown above.

"Mary had a little lamb" will be printed in Times Roman Standard font, 
11 point size, and so will the next line.  A font stays in effect for all
following text until a new font code appears.

The last line, "To be continued..." is printed in Times Roman Italic, 11 point,
as specified by <f3.  The p450> means that the printhead is to jump to the
right 450 pixels and then start printing this field of text at that point. This
"P" function is more precise than the "s" function for positioning the
beginning of printing at a specific place on the line. BUT, there is a
problem - the "p" code assumes that the printer is set at 300 DPI. Many 
printers are set at a higher DPI, like 1200 DPI on some laser printers. In
that case, telling the printhead to start printing at 450 pixels will cause
printing to start at about 1/4 the distance that it would print if 300 DPI
was used. The solution is a multiplier code, pulled from the application's
.INI file that tells what the "p" number should be multiplied by in order
to get the correct position on the printer being used.  So if 1200 DPI is
used, the multiplier would be 4.  Near the top of the ::SayMemo() method,
the .INI file is queried to get the nMultPoint factor, which is the multiplier
to compensate for a greater DPI.

GRAPHICS CODES - Boxes, Lines, Shades and Bitmaps

This is the really fun part! Changing fonts appropriately from one section
of text to another can greatly improve the look of your report. In addition,
adding boxes, shades, lines and bitmaps can really dress it up nicely. Codes
for doing this use the same kind of < > delimeters as the font codes, and they 
are placed right after the font code at the beginning of the field to which
the code applies, as shown in the example below.

An important concept to remember, is that before printing a page, the program
first lays out all the graphics for a line in a memory buffer, and then lays
down the text on top of the graphics.  That is why we have to specify the
graphic element used with a text field BEFORE we write the text of the field,
just as we have to specify the font for the field before printing it.

An interesting problem is how to deal with placing a box around variable length
text fields - like text coming from variable length memo fields. The graphic
codes for the box, with its dimensions, must be specified BEFORE printing the
field, yet we won't know where the bottom line of the box should be until
we finish printing the text, which may extend several lines down the page. 
This problem is solved by placing a <e> code ("extension") at the beginning of
each line of text from the variable field. When the program sees this <e> code,
it updates the nBottom dimension of the last box defined prior to this
<e> code.  It keeps updating the nBot dimension as each line is printed,
until the last line is laid down, and then that update becomes the final
nBot dimension that determines the bottom of the box.  The same principle
applies to the printing of vertical lines, except that an <h> code ("height")
is used to trigger the updating of the nBot dimension of the line.

However, if a variable length area that is to be boxed or have a vertical line
continues from the bottom of one page to the top of the next, then weird
things will happen unless the box and/or line is ended at the bottom of the
first page and then restarted at the top of the next page. This is
accomplished by using two static variables, lV := .t. and cGraphics := "<box &
line specs>". When the EndP() procedure is called in the report program to
check if the end of the page has been reached, if it finds a new page is to be
started, and if both lBoxes and lV are true, then <e><h> codes are inserted
at the end of the line just printed in order to make the printer end the
box and line at the bottom of the first page for the variable area. Then the
HeadP() procedure, which prints the header at the top of the next page,
refers to the lV and cGraphics variables to insert the box and line specs
at the beginning of the first line of text of the new page, in order to
restart the box and/or line for the variable area as it continues on
the new page.

Before we look at the sample below, one more critical item: you must place
a <t> code somewhere on the first line of text of your report to let REPT23
know that you are using graphic codes in this report. If it doesn't see this
<t> code, then REPT23 won't do the necessary setup step needed for printing
graphics. However, if it does see this code, then it executes the BoxSpecs()
setup function. This function, executed at the beginning of REPT23, scans the
entire text of the report for graphics codes, and then loads them into a
multi-dimension array for the program's reference during printing. Each row
of the array is keyed by page and line number so that the program can find
the graphic specs that apply to each line as it prints each line.  As each
line is printed, BoxWork() is executed to take care of implementing the
graphics, with reference to the ::aBoxSpecs array created by the BoxSpecs()
function.

There are 7 graphic codes:
   b  ... print a box
   c  ... print a box filled in with a shading
   s  ... print a shaded area without a box outline
   l  ... print a line
   e  ... extend the bottom of box
   h  ... (height) extend the bottom of a vertical line
   g  ... (graphic) print a bitmap

The syntax for a graphic code is as follows:

   <cGraphicCode nTop,nLft,nBot,nRt,nLinewidth,cBitmapName>
   
nTop, nLft, nBot and nRt are positional parameters that are measured
in inches. nLft and nRt are measured horizontally from the left edge of page, 
and nTop and nBot are measured vertically from the bottom edge of the line - 
where an underline character would be in the line.

During processing, the program automatically calculates the vertical position
from the top edge of the page for each nTop and nBot parameter by adding in
the top page margin plus the distance down from the top of the page based
on what line we are in and the assumption that we are using 6 vertical lines
to the inch.

Here is an example of using the graphic codes:

   <t><g -.27,.83,0,1.3,1,,orglogo.bmp>

   <f2><b -.2,.93,0,7.2,1,,>My Favorite Poem... <f3 s4>by Anonymous Artist

   <f1>Mary had a little lamb
            whose fleece were white as snow.

            <f3 p450>To be continued...

*/